# retinaface_pytorch

MagicMind 是面向寒武纪 MLU 的推理加速引擎。MagicMind 能将 AI 框架(Tensorflow,PyTorch,ONNX 等) 训练好的算法模型转换成 MagicMind 统一计算图表示,并提供端到端的模型优化、代码生成以及推理业务部署能力。

本 sample 探讨如何使用将 retinaface 网络的 PyTorch 实现转换为 MagicMind 模型，进而部署在寒武纪 MLU 板卡上。

## 目录

- [模型概述](#1模型概述)
- [前提条件](#2前提条件)
- [快速使用](#3快速使用)
  - [环境准备](#31-环境准备)
  - [下载仓库](#32-下载仓库)
  - [准备数据集](#33-准备数据集和模型)
  - [编译 MagicMind 模型](#34-编译-magicmind-模型)
  - [执行推理](#35-执行推理)
  - [一键运行](#36-一键运行)
- [高级说明](#4高级说明)
  - [export_model 高级说明](#41-export_model-高级说明)
  - [gen_model 高级说明](#42-gen_model-高级说明)
  - [infer_cpp 高级说明](#43-infer_cpp-高级说明)
- [精度和性能 benchmark](#5精度和性能-benchmark)
  - [性能 benchmark 测试](#51-性能-benchmark-测试)
  - [精度 benchmark 测试](#52-精度-benchmark-测试)
- [免责声明](#6免责声明)

## 1.模型概述

本例使用的 retinaface 实现来自 github 开源项目[https://github.com/biubug6/Pytorch_Retinaface](https://github.com/biubug6/Pytorch_Retinaface)，获取 master 分支最新代码(commit: b984b4b775b2c4dced95c1eadd195a5c7d32a60b)。下面将展示如何将该项目中 PyTorch 实现的 retinaface 模型转换为 MagicMind 的模型。

## 2.前提条件

请移至[主页面 README.md](../../../../README.md)的`2.前提条件`

## 3.快速使用

### 3.1 环境准备

请移至[主页面 README.md](../../../../README.md)的`3.环境准备`

### 3.2 下载仓库

```bash
# 下载仓库
git clone 本仓库
cd magicmind_cloud/buildin/cv/detection/retinaface_pytorch
```

在开始运行代码前需要执行以下命令安装必要的库：

```bash
pip install -r requirements.txt -i https://pypi.tuna.tsinghua.edu.cn/simple/ -f https://download.pytorch.org/whl/torch_stable.html
```

在开始运行代码前需要先检查 env.sh 里的环境变量，根据数据集实际路径修改 `env.sh` 内的 `WIDERFACE_DATASETS_PATH`,并且执行以下命令：

```bash
source env.sh
```

### 3.3 准备数据集和模型

```bash
cd ${PROJ_ROOT_PATH}/export_model
# bash run.sh <batch_size>
bash run.sh 1
```

### 3.4 编译 MagicMind 模型

```bash
cd ${PROJ_ROOT_PATH}/gen_model
# bash run.sh <magicmind_model> <precision> <batch_size> <dynamic_shape>
# 指定您想输出的magicmind_model路径，例如./model
bash run.sh ${magicmind_model} force_float32 1 true
```

### 3.5 执行推理

推理：

```bash
cd ${PROJ_ROOT_PATH}/infer_cpp
#bash runs.h <magicmind_model> <batch_size> 
bash run.sh qint8_mixed_float16 true 1 
```

结果：

```bash
==================== Results ====================
Easy   Val AP: 0.8954122458277102
Medium Val AP: 0.8646323548616941
Hard   Val AP: 0.6980163131947055
=================================================
```

### 3.6 一键运行

以上 3.3~3.6 的步骤也可以通过运行 `cd magicmind_cloud/buildin/cv/detection/retinaface_pytorch && bash run.sh` 来实现一键执行

## 4.高级说明

### 4.1 export_model 高级说明

修改 leaky 初始化为浮点数 0.0

```bash
leaky = 0.0
```

或者直接运行以下代码：

```bash
cd ${PROJ_ROOT_PATH}/export_model
# 修改Pytorch_Retinaface/models/net.py
git apply retinaface.diff
```

使用下面的代码导出 jit.trace 模型文件。

```bash
cd ${PROJ_ROOT_PATH}/export_model/Pytorch_Retinaface
python ../export.py --weights ${PROJ_ROOT_PATH}/data/models/Mobilenet0.25_Final.pth --imgsz 672 1024 --batch-size 1
```

### 4.2 gen_model 高级说明

PyTorch retinaface 模型转换为 MagicMind retinaface 模型分成以下几步：

- 使用 MagicMind Parser 模块将 torch.jit.trace 生成的 pt 文件解析为 MagicMind 网络结构。
- 模型量化。
- 使用 MagicMind Builder 模块生成 MagicMind 模型实例并保存为离线模型文件。

参数说明:

- `image_dir`: 输入图像目录，程序对该目录下所有后缀为 jpg 的图片执行目标检测任务。
- `conf_thres`: confidence_thresh，检测框得分阈值。
- `iou_thres`: nms_thresh。
- `max_det`: limit_detections。

**注意：**
在gen_model内使用了一些公共的组件，例如arg解析、第三方框架（如PyTorch）模型解析、MagicMind 配置设定等，这些公共组件的说明详见：[公共组件的README.md](../../../python_common/README.md)

### 4.3 infer_cpp 高级说明

概述：
本例使用 MagicMind C++ API 编写了名为 infer_cpp 的目标检测程序。infer_cpp 将展示如何使用 MagicMind C++ API 构建高效的 retinaface 人脸检测(图像预处理=>推理=>后处理)。其中程序主要由以下内容构成:

- infer.hpp, infer.cpp: 高效率的将 MagicMind 模型运行在 MLU 板卡上。
- pre_precess.hpp, pre_precess.cpp: 前处理。
- post_precess.hpp, post_precess.cpp: 后处理。

参数说明:

- `magicmind_model`: MagicMind 模型路径。
- `image_dir`: 输入图像目录，程序对该目录下所有后缀为 jpg 的图片执行目标检测任务。
- `image_num`: 输入图像数量。
- `file_list`: 数据集文件列表文件。
- `label_path`: 标签文件路径。
- `output_dir`: 根据检测结果进行渲染后的图像或检测结果 txt 文件保存路径。
- `save_img`: 结果可视化。若指定为 true，则保存渲染后的图像，默认为 false。
- `batch_size`: 推理时，输入 tensor 的维度

## 5.精度和性能 benchmark

### 5.1 性能 benchmark 测试

本仓库通过寒武纪提供的 Magicmind 性能测试工具 mm_run 展示性能数据

```bash
#查看参数说明
mm_run -h
mm_run --magicmind_model ${MM_MODEL} --batch_size ${BATCH_SIZE} --devices ${DEV_ID} --threads 1 --iterations 1000
```

或者通过一键运行 benchmark 里的脚本：

```bash
cd ${PROJ_ROOT_PATH}
bash benchmark/perf.sh
```

### 5.2 精度 benchmark 测试

一键运行 benchmark 里的脚本：

```bash
cd ${PROJ_ROOT_PATH}
bash benchmark/eval.sh
```

通过快速使用中 3.6 的脚本跑出 retinaface 在 widerface 数据集上的 AP 如下：
| Model | Precision | datasets | AP |
| ------------- | ------------- | ------------- | ------------- | 
| retinaface | force_float32 | Easy | 0.8991 |
| retinaface | force_float32 | Medium | 0.8674 | 
| retinaface | force_float32 | Hard | 0.7031 | 
| retinaface | force_float16 | Easy | 0.8991 | 
| retinaface | force_float16 | Medium | 0.8674 |
| retinaface | force_float16 | Hard | 0.7027 |
| retinaface | qint8_mixed_float16 | Easy | 0.8954 |
| retinaface | qint8_mixed_float16 | Medium | 0.8646 | 
| retinaface | qint8_mixed_float16 | Hard | 0.6980 | 


## 6.免责声明

您明确了解并同意，以下链接中的软件、数据或者模型由第三方提供并负责维护。在以下链接中出现的任何第三方的名称、商标、标识、产品或服务并不构成明示或暗示与该第三方或其软件、数据或模型的相关背书、担保或推荐行为。您进一步了解并同意，使用任何第三方软件、数据或者模型，包括您提供的任何信息或个人数据（不论是有意或无意地），应受相关使用条款、许可协议、隐私政策或其他此类协议的约束。因此，使用链接中的软件、数据或者模型可能导致的所有风险将由您自行承担。

- WIDER_val 数据集下载链接：[https://drive.google.com/uc?id=1GUCogbp16PMGa39thoMMeWxp7Rp5oM8Q](https://drive.google.com/uc?id=1GUCogbp16PMGa39thoMMeWxp7Rp5oM8Q)
- retinaface 模型下载链接：[https://drive.google.com/uc?id=14KX6VqF69MdSPk3Tr9PlDYbq7ArpdNUW](https://drive.google.com/uc?id=14KX6VqF69MdSPk3Tr9PlDYbq7ArpdNUW)
- retinaface GITHUB 下载链接：[https://github.com/biubug6/Pytorch_Retinaface.git](https://github.com/biubug6/Pytorch_Retinaface.git)

## 7.Release_Notes

@TODO
